@@ -10,9 +10,9 @@
  *               Iptc : https://www.iptc.org/std/IIM/4.1/specification/IIMV4.1.pdf
  *               Xmp  : https://www.adobe.com/devnet/xmp/pdfs/xmp_specification.pdf
  *                      https://www.iptc.org/std/Iptc4xmpCore/1.0/specification/Iptc4xmpCore_1.0-spec-XMPSchema_8.pdf
- *               Paper: http://www.metadataworkinggroup.com/pdf/mwg_guidance.pdf
+ *               Paper: www.metadataworkinggroup.com/pdf/mwg_guidance.pdf
  *
- * Copyright (C) 2006-2020 by Gilles Caulier <caulier dot gilles at gmail dot com>
+ * Copyright (C) 2006-2021 by Gilles Caulier <caulier dot gilles at gmail dot com>
  * Copyright (C) 2006-2013 by Marcel Wiesweg <marcel dot wiesweg at gmx dot de>
  *
  * This program is free software; you can redistribute it
@@ -52,6 +52,8 @@
 namespace Digikam
 {
 
+// TODO: merge with DMetadata class.
+
 class DIGIKAM_EXPORT MetaEngine
 {
 
@@ -115,6 +117,19 @@ public:
         ArraySeqTag  = 4
     };
 
+    /**
+     * Metadata Backend used to populate information.
+     */
+    enum Backend
+    {
+        Exiv2Backend       = 0,   ///< Default backend used by MetaEngine
+        LibRawBackend,            ///< DMetadata only
+        LibHeifBackend,           ///< DMetadata only
+        ImageMagickBackend,       ///< DMetadata only
+        FFMpegBackend,            ///< DMetadata only
+        NoBackend                 ///< No backend used (aka file cannot be read)
+    };
+
     /**
      * A map used to store Tags Key and Tags Value.
      */
@@ -142,11 +157,6 @@ public:
      */
     MetaEngine();
 
-    /**
-     * Copy constructor.
-     */
-    MetaEngine(const MetaEngine& metadata);
-
     /**
      * Constructor to load from parsed data.
      */
@@ -162,11 +172,6 @@ public:
      */
     virtual ~MetaEngine();
 
-    /**
-     * Create a copy of container
-     */
-    MetaEngine& operator=(const MetaEngine& metadata);
-
 public:
 
     //-----------------------------------------------------------------
@@ -176,28 +181,28 @@ public:
     /**
      * Return true if Exiv2 library initialization is done properly.
      * This method must be called before using libMetaEngine with multithreading.
-     * It initialize several non re-entrancy code from Adobe XMP SDK
-     * See Bug #166424 for details. Call cleanupExiv2() to clean things up later.
+     * It initialize several non re-entrancy code from Adobe XMP SDK, and register
+     * a function to cleanup automatically all XMP SDK memory allocation.
+     * See Bug #166424 for details.
      */
     static bool initializeExiv2();
 
     /**
-     * Return true if Exiv2 library memory allocations are cleaned properly.
-     * This method must be called after using libMetaEngine with multithreading.
-     * It cleans up memory used by Adobe XMP SDK
-     * See Bug #166424 for details.
-     */
-    static bool cleanupExiv2();
-
-    /**
-     * Return true if library can handle Xmp metadata
+     * Return true if Exiv2 library is compiled with Xmp metadata support.
      */
     static bool supportXmp();
 
+    /**
+     * Return true if library support Base Media File Format (aka CR3, HEIF, HEIC, and AVIF).
+     * Note: use this function only after to call initializeExiv2(), else false will aways returned.
+     * The function return true only if Exiv2 >= 0.27.4 compiled with BMFF support.
+     */
+    static bool supportBmff();
+
     /**
      * Return true if library can write metadata to typeMime file format.
      */
-    static bool supportMetadataWritting(const QString& typeMime);
+    static bool supportMetadataWriting(const QString& typeMime);
 
     /**
      * Return a string version of Exiv2 release in format "major.minor.patch"
@@ -351,11 +356,19 @@ public:
      */
     static bool hasSidecar(const QString& path);
 
+    /**
+     * Return a string of backend name used to parse metadata from file.
+     * See Backend enum for details.
+     */
+    static QString backendName(Backend t);
+
     /**
      * Load all metadata (Exif, Iptc, Xmp, and JFIF Comments) from a picture (JPEG, RAW, TIFF, PNG,
      * DNG, etc...). Return true if metadata have been loaded successfully from file.
+     * If backend is non null, return the backend used to populate metadata (Exiv2).
+     * See Backend enum for details.
      */
-    virtual bool load(const QString& filePath);
+    bool load(const QString& filePath, Backend* backend = nullptr);
 
     /**
      * Load metadata from a sidecar file and merge.
@@ -375,6 +388,14 @@ public:
      */
     bool applyChanges(bool setVersion = false) const;
 
+    /**
+     * List changed tags compared to original file contents and export result to
+     * a temporary EXV file container.
+     * 'exvTmpFile' is the path to the temporary EXV container to create.
+     * 'removedTags' is populated with the list of tags removed.
+     */
+    bool exportChanges(const QString& exvTmpFile, QStringList& removedTags) const;
+
     //@}
 
     //-------------------------------------------------------------------
@@ -433,14 +454,14 @@ public:
      * Set the Exif and Iptc time stamp. If 'setDateTimeDigitized' parameter is true, the 'Digitalized'
      * time stamp is set, else only 'Created' time stamp is set.
      */
-    bool setImageDateTime(const QDateTime& dateTime, bool setDateTimeDigitized=false) const;
+    bool setImageDateTime(const QDateTime& dateTime, bool setDateTimeDigitized = false) const;
 
     /**
      * Return the digitization time stamp of the item. First Exif information is checked, then IPTC.
      * If no digitization time stamp is found, getItemDateTime() is called if fallbackToCreationTime
      * is true, or a null QDateTime is returned if fallbackToCreationTime is false.
      */
-    QDateTime getDigitizationDateTime(bool fallbackToCreationTime=false) const;
+    QDateTime getDigitizationDateTime(bool fallbackToCreationTime = false) const;
 
     /**
      * Return a QImage copy of Iptc preview image. Return a null item if preview cannot
@@ -455,7 +476,7 @@ public:
      * Re-implement this method if you want to use another item file format than JPEG to
      * save preview.
      */
-    virtual bool setItemPreview(const QImage& preview) const;
+    bool setItemPreview(const QImage& preview) const;
 
     //@}
 
@@ -544,7 +565,7 @@ public:
      * Set true 'addExifHeader' parameter to add an Exif header to Exif metadata.
      * Returns a null Qt byte array if there is no Exif metadata in memory.
      */
-    QByteArray getExifEncoded(bool addExifHeader=false) const;
+    QByteArray getExifEncoded(bool addExifHeader = false) const;
 
     /**
      * Set the Exif data using a Qt byte array. Return true if Exif metadata
@@ -597,7 +618,7 @@ public:
      * Get an Exif tags content like a string. If 'escapeCR' parameter is true, the CR characters
      * will be removed. If Exif tag cannot be found a null string is returned.
      */
-    QString getExifTagString(const char* exifTagName, bool escapeCR=true) const;
+    QString getExifTagString(const char* exifTagName, bool escapeCR = true) const;
 
     /**
      * Set an Exif tag content using a string. Return true if tag is set successfully.
@@ -624,7 +645,7 @@ public:
      * 'num' and 'den' are the numerator and the denominator of the rational value.
      * Return true if Exif tag be found.
      */
-    bool getExifTagRational(const char* exifTagName, long int& num, long int& den, int component=0) const;
+    bool getExifTagRational(const char* exifTagName, long int& num, long int& den, int component = 0) const;
 
     /**
      * Set an Exif tag content using a rational value.
@@ -653,7 +674,8 @@ public:
      * if rationalAsListOfInts is true, as double if rationalAsListOfInts is false.
      * An exif tag of numerical type may contain more than one value; set component to the desired index.
      */
-    QVariant getExifTagVariant(const char* exifTagName, bool rationalAsListOfInts=true, bool escapeCR=true, int component=0) const;
+    QVariant getExifTagVariant(const char* exifTagName, bool rationalAsListOfInts = true,
+                               bool escapeCR = true, int component = 0) const;
 
     /**
      * Set an Exif tag content using a QVariant. Returns true if tag is set successfully.
@@ -663,7 +685,7 @@ public:
      * Setting a value with multiple components is currently not supported.
      */
     bool setExifTagVariant(const char* exifTagName, const QVariant& data,
-                           bool rationalWantSmallDenominator=true) const;
+                           bool rationalWantSmallDenominator = true) const;
 
     /**
      * Remove the Exif tag 'exifTagName' from Exif metadata. Return true if tag is
@@ -687,7 +709,7 @@ public:
      * This is equivalent to calling getExifTagString directly.
      * If escapeCR is true CR characters will be removed from the result.
      */
-    QString createExifUserStringFromValue(const char* exifTagName, const QVariant& val, bool escapeCR=true);
+    QString createExifUserStringFromValue(const char* exifTagName, const QVariant& val, bool escapeCR = true);
 
     /**
      * Return a map of Exif tags name/value found in metadata sorted by
@@ -709,8 +731,13 @@ public:
      *   if 'inverSelection' is false.
      * - not include "Iop", or "Thumbnail", or "Image", or "Photo" in the Exif tag keys
      * if 'inverSelection' is true.
+     * if 'extractBinary" is true, tags with undefined types of data are extracted (default),
+     * else contents is replaced by "Binary data ... bytes". Take a care as large binary data as
+     * origianl RAW data from DNG container can be huge and listing Exif tags from GUI can take a while.
      */
-    MetaEngine::MetaDataMap getExifTagsDataList(const QStringList& exifKeysFilter=QStringList(), bool invertSelection=false) const;
+    MetaEngine::MetaDataMap getExifTagsDataList(const QStringList& exifKeysFilter = QStringList(),
+                                                bool invertSelection = false,
+                                                bool extractBinary = true) const;
 
     //@}
 
@@ -743,7 +770,7 @@ public:
      * Set true 'addIrbHeader' parameter to add an Irb header to Iptc metadata.
      * Return a null Qt byte array if there is no Iptc metadata in memory.
      */
-    QByteArray  getIptc(bool addIrbHeader=false) const;
+    QByteArray  getIptc(bool addIrbHeader = false) const;
 
     /**
      * Set the Iptc data using a Qt byte array. Return true if Iptc metadata
@@ -755,7 +782,7 @@ public:
      * Get an Iptc tag content like a string. If 'escapeCR' parameter is true, the CR characters
      * will be removed. If Iptc tag cannot be found a null string is returned.
      */
-    QString getIptcTagString(const char* iptcTagName, bool escapeCR=true) const;
+    QString getIptcTagString(const char* iptcTagName, bool escapeCR = true) const;
 
     /**
      * Set an Iptc tag content using a string. Return true if tag is set successfully.
@@ -770,7 +797,7 @@ public:
      * will be removed.
      * If no tag can be found an empty list is returned.
      */
-    QStringList getIptcTagsStringList(const char* iptcTagName, bool escapeCR=true) const;
+    QStringList getIptcTagsStringList(const char* iptcTagName, bool escapeCR = true) const;
 
     /**
      * Set multiple Iptc tags contents using a strings list. 'maxSize' is the max characters size
@@ -825,7 +852,8 @@ public:
      * - not include "Envelope", or "Application2" in the Iptc tag keys
      *   if 'inverSelection' is true.
      */
-    MetaEngine::MetaDataMap getIptcTagsDataList(const QStringList& iptcKeysFilter=QStringList(), bool invertSelection=false) const;
+    MetaEngine::MetaDataMap getIptcTagsDataList(const QStringList& iptcKeysFilter = QStringList(),
+                                                bool invertSelection = false) const;
 
     /**
      * Return a strings list of Iptc keywords from item. Return an empty list if no keyword are set.
@@ -909,7 +937,7 @@ public:
      * Get a Xmp tag content like a string. If 'escapeCR' parameter is true, the CR characters
      * will be removed. If Xmp tag cannot be found a null string is returned.
      */
-    QString getXmpTagString(const char* xmpTagName, bool escapeCR=true) const;
+    QString getXmpTagString(const char* xmpTagName, bool escapeCR = true) const;
 
     /**
      * Set a Xmp tag content using a string. Return true if tag is set successfully.
@@ -953,8 +981,8 @@ public:
      * - not include "dc", or "xmp" in the Xmp tag keys
      *   if 'inverSelection' is true.
      */
-    MetaEngine::MetaDataMap getXmpTagsDataList(const QStringList& xmpKeysFilter=QStringList(),
-                                               bool invertSelection=false) const;
+    MetaEngine::MetaDataMap getXmpTagsDataList(const QStringList& xmpKeysFilter = QStringList(),
+                                               bool invertSelection = false) const;
 
     /**
      * Get all redondant Alternative Language Xmp tags content like a map.
@@ -963,7 +991,7 @@ public:
      * If Xmp tag cannot be found a null string list is returned.
      */
     MetaEngine::AltLangMap getXmpTagStringListLangAlt(const char* xmpTagName,
-                                                      bool escapeCR=true) const;
+                                                      bool escapeCR = true) const;
 
     /**
      * Set an Alternative Language Xmp tag content using a map. See AltLangMap class
@@ -993,7 +1021,7 @@ public:
      * Get a Xmp tag content like a sequence of strings. If 'escapeCR' parameter is true, the CR characters
      * will be removed from strings. If Xmp tag cannot be found a null string list is returned.
      */
-    QStringList getXmpTagStringSeq(const char* xmpTagName, bool escapeCR=true) const;
+    QStringList getXmpTagStringSeq(const char* xmpTagName, bool escapeCR = true) const;
 
     /**
      * Set a Xmp tag content using the sequence of strings 'seq'.
@@ -1039,7 +1067,7 @@ public:
      * LangAlt values will have type Map (QMap<QString, QVariant>) with the language
      * code as key and the contents as value, of type String.
      */
-    QVariant getXmpTagVariant(const char* xmpTagName, bool rationalAsListOfInts=true, bool stringEscapeCR=true) const;
+    QVariant getXmpTagVariant(const char* xmpTagName, bool rationalAsListOfInts = true, bool stringEscapeCR = true) const;
 
     /**
      * Return a strings list of Xmp keywords from item. Return an empty list if no keyword are set.
@@ -1103,12 +1131,11 @@ public:
      * Remove the Xmp tag 'xmpTagName' from Xmp metadata. Return true if tag is
      * removed successfully or if no tag was present.
      */
-    bool removeXmpTag(const char* xmpTagName) const;
-
+    bool removeXmpTag(const char* xmpTagName, bool family = false) const;
 
     /**
      * Register a namespace which Exiv2 doesn't know yet. This is only needed
-     * when new Xmp properties are added manually. 'uri' is the namespace url and prefix the
+     * when new Xmp properties are added manually. 'uri' is the namespace url and 'prefix' the
      * string used to construct new Xmp key (ex. "Xmp.digiKam.tagList").
      * NOTE: If the Xmp metadata is read from an item, namespaces are decoded and registered
      * by Exiv2 at the same time.
@@ -1140,7 +1167,7 @@ public:
      * Get GPS location information set in the item, in the GPSCoordinate format
      * as described in the XMP specification. Returns a null string in the information cannot be found.
      */
-    QString getGPSLatitudeString() const;
+    QString getGPSLatitudeString()  const;
     QString getGPSLongitudeString() const;
 
     /**
@@ -1148,7 +1175,7 @@ public:
      * where the sign determines the direction ref (North + / South - ; East + / West -).
      * Returns true if the information is available.
      */
-    bool getGPSLatitudeNumber(double* const latitude) const;
+    bool getGPSLatitudeNumber(double* const latitude)   const;
     bool getGPSLongitudeNumber(double* const longitude) const;
 
     /**
@@ -1204,7 +1231,13 @@ public:
                                                   long int* const denominator);
 
     /**
-     *Converts a GPS position stored as rationals in Exif to the form described
+     * Converts degrees values as a double representation. This code take a care about hemisphere position.
+     */
+    static double convertDegreeAngleToDouble(double degrees, double minutes, double seconds);
+
+
+    /**
+     * Converts a GPS position stored as rationals in Exif to the form described
      * as GPSCoordinate in the XMP specification, either in the from "256,45,34N" or "256,45.566667N"
      */
     static QString convertToGPSCoordinateString(const long int numeratorDegrees,
@@ -1212,7 +1245,7 @@ public:
                                                 const long int numeratorMinutes,
                                                 const long int denominatorMinutes,
                                                 const long int numeratorSeconds,
-                                                long int denominatorSeconds,
+                                                const long int denominatorSeconds,
                                                 const char directionReference);
 
     /**
@@ -1277,6 +1310,17 @@ protected:
      */
     bool setProgramId() const;
 
+private:
+
+    // Disable copy constructor and operator to prevent potential slicing with this class, reported by Clazy static analyzer.
+    // https://github.com/KDE/clazy/blob/master/docs/checks/README-copyable-polymorphic.md
+    // This methods was implemented to be able to pass this class or a derived version to signals and slots. This is very
+    // Dangerous as virtual methods are present in this polymorphic class and is copyable.
+    // Instead to use this class in signals and slots, use MetaEngineData container.
+    // TODO: remove legacy implementations for these methods later if no side effect.
+    MetaEngine(const MetaEngine& metadata);
+    MetaEngine& operator=(const MetaEngine& metadata);
+
 private:
 
     /**
